# pylint: disable=no-member, invalid-name
""" Read ICP data (xlsx) """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
import re
import matplotlib.pyplot as plt
import pandas as pd


def load(fstr, resample=None):
    """
    load .xlsx file for ICP using pandas
    average all time stamp (resampling) in 'seconds' (default)
    return a new time series by pandas
    """
    # open .xlsx file
    xl_file = pd.ExcelFile(fstr)

    # parse only the first sheet (xlsx file has many sheets)
    df = xl_file.parse(xl_file.sheet_names[0])

    # using values to extract data from DataFrame
    icp_data = df["p"].values
    icp_timestmp = df["timestmp"].values
    # print('file %s size is %d' % (fstr, icp_data.size))

    # build a timeseries structure
    ts = pd.Series(icp_data, index=icp_timestmp)

    if resample is not None:
        # resample in 'seconds' as EIT system data rate
        ts = ts.resample(resample).mean()

    # return pandas series, you may use ts.index to extract time stamp
    # beware, there are np.NaN values in tss !
    return ts


# load in .csv format, much faster
def load_csv(fstr, resample=None):
    """
    load .xlsx file for ICP using pandas
    average all time stamp (resampling) in 'seconds' (default)

    Returns
    -------
    a new time series by pandas

    Notes
    -----
    to_datetime : ISO standard time format is XXXX-XX-XX DD:DD:DD.DDDD
    so as the format differs, loading performance drops.

    We may specify format,
    >>> t = pd.to_datetime(df['timestmp'], format="%Y/%m/%d %H:%M:%S.%f")
    however this is still slow before pandas-0.17.1

    instead, we apply a function row-wise on a series using map in DataFrame
    (However, you should really avoid apply with custom Python functions)
    it can work and works fine.
    """
    df = pd.read_csv(fstr)

    def f(x):
        """f = lambda x: re.sub('/', '-', x)"""
        return re.sub("/", "-", x)

    timestr = df["timestmp"].map(f)
    icp_timestmp = pd.to_datetime(timestr)
    icp_data = df["p"].values

    # build a timeseries structure
    ts = pd.Series(icp_data, index=icp_timestmp)

    if resample is not None:
        # averaged over 'seconds' as EIT systems
        ts = ts.resample(resample).mean()

    return ts


def convert(file_from, file_to, resample="s"):
    """
    load icp from xlsx and save it to a csv file,
    resample by default
    """
    if file_from.find("xlsx") > 0:
        ts = load(file_from, resample=resample)
    else:
        ts = load_csv(file_from, resample=resample)

    ts.to_csv(file_to)
    return ts


def demo_read_xlsx(filestr):
    """
    demo on how to load original (.xlsx) file
    """
    ts = load(filestr)
    return ts


def demo_read_csv(filestr):
    """
    export .xlsx in .csv and read using load_csv(), it is much faster
    """
    ts = load_csv(filestr, resample="600s")
    return ts


if __name__ == "__main__":
    ts_data = demo_read_csv("../../datasets/all.csv")

    # plot and see
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.plot(ts_data)
    ax.grid("on")
    plt.show()



# coding: utf-8
# pylint: disable=no-member, invalid-name, too-many-arguments
""" plot function based on vispy for tetrahedral plots """
# liubenyuan <liubenyuan@gmail.com>
# 2015, 2016, 2017
from __future__ import absolute_import

import sys
import numpy as np
import matplotlib
import matplotlib.cm as cm
from matplotlib.colors import LinearSegmentedColormap

#
from vispy import app, gloo, scene
from vispy.visuals import Visual

#
from pyeit.mesh.plot.simconv import sim2edge, sim2tri

# build vertex shader for tetplot
vert = """
uniform vec4 u_color;
attribute vec4 a_color;
varying vec4 v_color;

void main()
{
    vec4 visual_pos = vec4($position, 1);
    vec4 doc_pos = $visual_to_doc(visual_pos);
    gl_Position = $doc_to_render(doc_pos);

    v_color = a_color * u_color;
}
"""

# build fragment shader for tetplot
frag = """
varying vec4 v_color;

void main()
{
    gl_FragColor = v_color;
}
"""


class TetPlotVisual(Visual):
    """template"""

    def __init__(
        self,
        points,
        simplices,
        vertex_color=None,
        mask_color=None,
        alpha=1.0,
        mode="triangles",
    ):
        """initialize tetrahedra face plot

        Parameters
        ----------
        points : NDArray of float32
            N x 3 points coordinates
        simplices : NDArray of uint32
            N x 4 connectivity matrix

        Note
        ----
        initialize triangles structure
        """
        Visual.__init__(self, vcode=vert, fcode=frag)

        # set data
        self.shared_program.vert["position"] = gloo.VertexBuffer(points)
        if vertex_color is None:
            vertex_color = np.ones((points.shape[0], 4), dtype=np.float32)
        else:
            assert vertex_color.shape[0] == points.shape[0]
            # vertex color may be grayscale
            if np.ndim(vertex_color) == 1:
                f = vertex_color[:, np.newaxis]
                v = np.repeat(f, 4, axis=1)
                v[:, -1] = 1.0
                vertex_color = v.astype(np.float32)
        self.shared_program["a_color"] = vertex_color

        # mask colors, alpha channel is not used when mask_color is given.
        if mask_color is None:
            mask_color = [1.0, 1.0, 1.0, alpha]
        self.shared_program["u_color"] = mask_color

        # build buffer
        if mode == "triangles":
            vbo = sim2tri(simplices)
        elif mode == "lines":
            vbo = sim2edge(simplices)
        else:
            raise ValueError("Drawing mode = " + mode + " not supported")
        self._index_buffer = gloo.IndexBuffer(vbo)

        # config OpenGL, 'translucent' or 'additive'
        self.set_gl_state(
            "additive",
            blend=True,
            depth_test=False,
            cull_face=False,
            polygon_offset_fill=False,
            polygon_offset=(1, 1),
        )
        self._draw_mode = mode

    @staticmethod
    def _prepare_transforms(view):
        """This method is called when the user or the scenegraph has assigned
        new transforms to this visual"""
        # Note we use the "additive" GL blending settings so that we do not
        # have to sort the mesh triangles back-to-front before each draw.
        tr = view.transforms
        view_vert = view.view_program.vert
        view_vert["visual_to_doc"] = tr.get_transform("visual", "document")
        view_vert["doc_to_render"] = tr.get_transform("document", "render")


def tetplot(
    points, simplices, vertex_color=None, edge_color=None, alpha=1.0, axis=True
):
    """main function for tetplot"""
    TetPlot = scene.visuals.create_visual_node(TetPlotVisual)

    # convert data types for OpenGL
    pts_float32 = points.astype(np.float32)
    sim_uint32 = simplices.astype(np.uint32)

    # The real-things : plot using scene
    # build canvas
    canvas = scene.SceneCanvas(keys="interactive", show=True)

    # Add a ViewBox to let the user zoom/rotate
    view = canvas.central_widget.add_view()
    view.camera = "turntable"
    view.camera.fov = 50
    view.camera.distance = 3

    if vertex_color is not None and vertex_color.ndim == 1:
        vertex_color = blue_red_colormap(vertex_color)

    # drawing only triangles
    # 1. turn off mask_color, default = [1.0, 1.0, 1.0, alpha]
    # 2. mode = 'triangles'
    TetPlot(
        pts_float32,
        sim_uint32,
        vertex_color,
        mask_color=None,
        alpha=alpha,
        mode="triangles",
        parent=view.scene,
    )

    # drawing only lines
    # 1. turn off vertex_color, default = [[1.0, 1.0, 1.0, 1.0]*N]
    # 2. mode = 'lines'
    # 3. alpha channel is specified instead of mask_color
    if edge_color is not None:
        TetPlot(
            pts_float32,
            sim_uint32,
            vertex_color=None,
            mask_color=edge_color,
            alpha=alpha,
            mode="lines",
            parent=view.scene,
        )

    # show axis
    if axis:
        scene.visuals.XYZAxis(parent=view.scene)

    # run
    app.run()


def blue_red_colormap(f):
    """mapping vector to blue (-) red (+) color map"""
    # convert vertex_color
    cdict1 = {
        "red": ((0.0, 0.0, 0.0), (0.5, 0.0, 0.1), (1.0, 1.0, 1.0)),
        "green": ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0)),
        "blue": ((0.0, 1.0, 1.0), (0.5, 0.1, 0.0), (1.0, 0.0, 0.0)),
    }
    cdict1["alpha"] = (
        (0.00, 1.0, 1.0),
        (0.25, 0.6, 0.6),
        (0.50, 0.0, 0.0),
        (0.75, 0.6, 0.6),
        (1.00, 1.0, 1.0),
    )

    def blue_red():
        """interpolate blue red color"""
        return LinearSegmentedColormap("BlueRed", cdict1)

    # map vector to RGBA
    maxima = np.max(np.abs(f))
    minima = -maxima
    brcmap = blue_red()
    norm = matplotlib.colors.Normalize(vmin=minima, vmax=maxima, clip=True)
    mapper = cm.ScalarMappable(norm=norm, cmap=brcmap)
    v = mapper.to_rgba(f)
    return v.astype(np.float32)


# demo
if __name__ == "__main__":
    if sys.flags.interactive != 1:
        # location of points
        pts = np.array(
            [
                (0.0, 0.0, 0.0),
                (1.0, 0.0, 0.0),
                (0.0, 1.0, 0.0),
                (0.0, 0.0, 1.0),
                (1.0, 1.0, 1.0),
            ],
            dtype=np.float32,
        )

        # connectivity of two tetrahedrons
        sim = np.array([(0, 1, 2, 3), (1, 3, 2, 4)], dtype=np.uint32)

        # plot
        tetplot(pts, sim, edge_color=[0.2, 0.2, 1.0, 0.2], alpha=0.1, axis=False)



# pylint: disable=no-member, invalid-name, too-many-instance-attributes
"""
load .et0, .et3, .erd file into mem
using pack and unpack together with regular expression to filter out data.

These file types were developed by FMMU EIT group.
Please cite the following paper if you are using et3 in your research:
    Fu, Feng, et al. "Use of electrical impedance tomography to monitor
    regional cerebral edema during clinical dehydration treatment."
    PloS one 9.12 (2014): e113202.
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
import os
from struct import unpack

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import dates


class ET3:
    """et3, erd file loader"""

    def __init__(
        self,
        file_name,
        data_type="auto",
        rel_date=None,
        fps=1,
        reindex=False,
        trim=False,
        verbose=False,
    ):
        """
        initialize file handler (supports .et0, .et3, .erd)

        Parameters
        ----------
        file_name : String
            file path for EIT data.
        data_type : String, optional
            manually set data type i.e., "et0". The default is "auto".
        rel_date : Datetime, optional
            Datetime, i.e., "2014/01/25". If specified, the time information
            in header is ignored. The default is None.
        fps : Int, optional
            Frame per second, valid when rel_date specified. The default is 1.
        reindex : Bool, optional
            reindex ERD data order to ET3 format (RALP -> LARP) before trim.
        trim : Bool, optional
            Measurements on current carry electrodes are discarded.
        verbose : Bool, optional
            Print debuging information. The default is False.
        """
        self.file_name = file_name
        self.ext = os.path.splitext(file_name)[1][1:].lower()
        self.rel_date = rel_date
        self.fps = fps
        self.reindex = reindex
        self.trim = trim
        self.verbose = verbose

        # frame = frame header + 2x256 (Re, Im) doubles frame data
        self.header_size = 1024  # Bytes
        self.n_data = 2 * 256  # doubles
        self.data_size = self.n_data * 8  # Bytes
        self.frame_size = self.header_size + self.data_size  # Bytes
        self.nadc = 8  # number of maximum ADC channels

        file_size = os.path.getsize(file_name)
        offset = file_size % self.frame_size
        # 'deprecated' file type has an extra file header (4096 Bytes)
        # 'et' : each frame is 5120 Bytes, no extra file header
        # 5120 = 1024 Bytes frame header + 256x2 Double frame data (4096 Bytes)
        if not ((offset == 0) or (offset == 4096)):
            raise ValueError("Wrong File Offset = {}".format(offset))
        n_frame = int((file_size - offset) / self.frame_size)
        self.offset = offset
        self.n_frame = n_frame

        # extract system configuration
        self.p = self.setup(data_type)
        # load data (RAW data are complex-valued) and build datetime
        time_array, self.data, self.adc_array = self.load()
        self.ts = self.build_ts(time_array)

    def setup(self, data_type):
        """
        Infer file-type and header information
        """
        # parse header
        parser = parse_header_et0 if data_type == "et0" else parse_header
        with open(self.file_name, "rb") as fh:
            fh.read(self.offset)  # skip offset
            d = fh.read(self.header_size)
            p = parser(d)  # extract information from frame header

        # default et3: version <= 3 and ext in ["et0", "et3"]
        data_format = "et"
        date0 = "1899/12/30"  # excel format
        ts_scale = 86400  # convert days to seconds
        ts_offset = 0
        scale = 1.0
        # other formats
        if data_type == "et0":  # et0: convert voltage to Ohm
            scale = gain_table(p["gain"], p["current"])
        if p["version"] == 4 and self.ext == "erd":
            data_format = "erd"
            date0 = "1970/01/01"  # erd: posix format
            ts_scale = 0.001
            ts_offset = 8 * 3600
            scale = 4096 / 65536 / 29.9 * 1000 / 1250  # fixed gain

        p_new = {
            "data_format": data_format,
            "date0": date0,
            "ts_scale": ts_scale,
            "ts_offset": ts_offset,
            "scale": scale,
        }
        p.update(p_new)

        return p

    @staticmethod
    def erd2et():
        """
        CT scans RALP, Starting at the 9 o’clock position and moving clockwise
        in 90 degree intervals, we are looking at the
        Right, Anterior, Left and Posterior aspects of the patient.

        ERD electrodes: 0 (R) 4 (A) 8 (L) 12 (P)
        ET electrodes : 0 (L) 4 (A) 8 (R) 12 (P)

        These systems are using opposition stimulation (a, b) = (0-8), (1-9), ..
        and rotate measurements a + (1-0, 2-1, .., 15-14).
        """
        erd_electrodes = np.arange(16)  # 0 .. 15
        et3_electrodes = np.hstack([np.arange(9, 0, -1), np.arange(16, 9, -1)]) - 1
        row_swap = [np.where(i == et3_electrodes)[0][0] for i in erd_electrodes]
        ind = np.arange(256).reshape(16, -1)
        reind = np.fliplr(ind[row_swap, :]).reshape(-1)  # CCW to CW
        return reind

    def load(self):
        """load frames (header + data)"""
        time_array = np.zeros(self.n_frame)
        x = np.zeros((self.n_frame, self.n_data), dtype=np.double)
        adc_array = np.zeros((self.n_frame, self.nadc), dtype=np.double)
        with open(self.file_name, "rb") as fh:
            fh.read(self.offset)
            for i in range(self.n_frame):
                # get a whole frame
                d = fh.read(self.frame_size)
                # extract time ticks
                time_array[i] = unpack("d", d[8:16])[0]
                # extract ADC samples
                dp = d[960 : self.header_size]
                adc_array[i] = np.array(unpack("8d", dp))
                # extract demodulated I,Q data
                x[i] = np.array(unpack("512d", d[self.header_size :]))

        # build complex-valued data from Re, Im measurements
        data = x[:, :256] + 1j * x[:, 256:]

        # reindex ERD format to FMMU ET format
        if self.p["data_format"] == "erd" and self.reindex is True:
            ind = self.erd2et()
            data = data[:, ind]

        # remove all measurements on current carring electrodes
        if self.trim:
            idx = trim_pattern()
            data = data[:, idx]

        # rescale data to Ohms
        data = data * self.p["scale"]

        return time_array, data, adc_array

    def build_ts(self, time_array):
        """convert delta time (seconds) to datetime series"""
        if self.rel_date is not None:  # user sets datetime
            rel_date = self.rel_date
            d_seconds = np.arange(self.n_frame) * 1.0 / self.fps
        else:
            rel_date = self.p["date0"]
            d_seconds = time_array * self.p["ts_scale"] + self.p["ts_offset"]

        d_seconds = np.round(d_seconds * 1000.0) / 1000.0  # 1 ms resolution
        ts = pd.to_datetime(rel_date) + pd.to_timedelta(d_seconds, "s")

        # check duplicated
        if any(ts.duplicated()):
            print("{}: duplicated datetime:".format(self.file_name))
            print(ts[ts.duplicated()])

        return ts

    def to_df(self):
        """convert raw EIT data to DataFrame"""
        df = pd.DataFrame(self.data, index=self.ts)
        df = df[~df.index.duplicated()]
        return df

    def to_dp(self, adc_filter=False):
        """convert raw ADC data to DataFrame"""
        # left ear, right ear, Nasopharyngeal, rectal
        columns = ["tle", "tre", "tn", "tr", "c4", "c5", "c6", "c7"]
        dp = pd.DataFrame(self.adc_array, index=self.ts, columns=columns)
        dp = dp[~dp.index.duplicated()]

        if adc_filter:
            # correct temperature (temperature can accidently be 0)
            dp.loc[dp["tle"] == 0, "tle"] = np.nan
            dp.loc[dp["tre"] == 0, "tre"] = np.nan
            dp.loc[dp["tn"] == 0, "tn"] = np.nan
            dp.loc[dp["tr"] == 0, "tr"] = np.nan

            # filter auxillary sampled data
            dp.tle = med_outlier(dp.tle)
            dp.tre = med_outlier(dp.tre)
            dp.tn = med_outlier(dp.tn)
            dp.tr = med_outlier(dp.tr)

        return dp

    def to_csv(self):
        """
        save data in a .csv file

        this function is embedded in to pandas.DataFrame, simply call
        $ df.to_csv(file_to, columns=None, header=False, index=False)
        """
        raise NotImplementedError


def med_outlier(d, window=17):
    """filter outliers using median filter"""
    med = d.rolling(window, center=False).median()
    std = d.rolling(window, center=False).std()
    std[std == np.nan] = 0.0
    # replace med with d for outlier removal
    df = med[(d <= med + 3 * std) & (d >= med - 3 * std)]
    return df


def parse_header(d):
    """
    Parse the header of et files, bytes order:

    version: int (4 Bytes)
    frame index: int (4 Bytes)
    time: double (8 bytes)
    tt: 3*double (reserved)
    -- offset 5 double (40 Bytes) --

    Reconstruction Parameters
    -- offset 40 double (320 Bytes) --

    struct configinfo{
        DWORD dwDriveDelay;     // driver switch interval (us)
        DWORD dwMeasureDelay;   // measure switch interval (us)
        int nDrvNode;           // the distance of two driving electrodes
        int nMeaNode;           // the distance of two measure electrodes
        int nFrequency;         // Hertz
        int nCurrent;           // uA
        int nGain;              // Gain number
        int nElecNum;           // number of electrodes
        float fElecSize;        // size of electrodes
        float fPeriod;          // frame interval (s)
    };
    measure info: 5*16 int
    -- offset 90 int (360 Bytes)

    electrode SQ: 16 double
    reserved 14 double
    -- offset 30 double (240 Bytes)

    temperature 6 channel (double)
    reserved 16 Bytes
    -- offset 8 double (64 Bytes)

    total header = 40 + 320 + 360 + 240 + 64 = 1024
    """
    # extract version info {et0: NA, et3: 1, shi: 3}
    version = int(unpack("I", d[:4])[0])
    h = np.array(unpack("8I2f", d[360:400]))
    frequency = h[4]
    current = h[5]
    gain = h[6]

    p = {
        "version": version,
        "frequency": frequency,
        "current": current,
        "gain": gain,
    }
    return p


def parse_header_et0(d):
    """
    parse et0 header. Guess from binary dump (byliu)
    binary dump all (little endian, i.e., LSB 16 bit first):

    print('now dump')
    h_all = np.array(unpack('256I', d))
    for i in range(32):
        h_seg = h_all[i*8 + np.arange(8)]
        print(','.join('{:02x}'.format(x) for x in h_seg))

    current can also be read out in each switches:

    header_offset = 84
    h = np.array(unpack('4H', d[header_offset:header_offset+8]))
    nGain, nCurrent = h[1], h[3]
    print(nGain, nCurrent)

    hints on extract date from .et0 file.
    for .et0 file date is stored in string (bytes) format (utf-16, CJK)
    i.e., '2016Y01M01D 11H08M30S', offset=4, length=42 (21*2) Bytes
    """
    h = np.array(unpack("8H", d[48:64]))
    # print(','.join('{:02x}'.format(x) for x in h))

    # extract information in global configurations
    frequency = np.int(h[1])
    current = np.int(h[3])
    gain = np.int(h[5])

    p = {
        "version": 0,  # .et0
        "frequency": frequency,
        "current": current,
        "gain": gain,
    }
    return p


def gain_table(gain, current_in_ua):
    """
    Rescale data using if EIT is using programmable gain.

    pg table (in dll) used by Zhang Ge, 2017/12/06
    pgia_table = {0: 0.08,
                  1: 0.16,
                  2: 0.32,
                  3: 0.63,
                  4: 1.26,
                  5: 2.52,
                  6: 5.01,
                  7: 10.0}
    gain = 25.7 * keyvalue
    """
    # Programmable Gain table (scale mapping), see MainFrm.cpp
    pgia_table = {
        0: 4.112,
        1: 8.224,
        2: 16.448,
        3: 32.382,
        4: 64.764,
        5: 129.528,
        6: 257.514,
        7: 514,
    }

    # make sure gain is a valid key
    if gain not in pgia_table.keys():
        gain = 3
    if current_in_ua <= 0 or current_in_ua > 1250:
        current_in_ua = 1000

    # mapping ADC to resistor
    voltage_in_uv = 2.5 * 1000000.0 / 32768.0 / pgia_table[gain]
    scale = voltage_in_uv / current_in_ua

    return scale


def trim_pattern():
    """
    Generate trim array (masked value, 0) on rotating measurements,
    0......00......0 0......00......0 .. 0......00......0
    """
    idx = np.ones(256, dtype=np.bool)

    for i in range(32):
        j = i * 8
        # exclude the stimulation indices (i.e., 0-8 stimulus)
        idx[j] = False
        idx[j + 7] = False

    return idx


if __name__ == "__main__":
    file_name = "/data/dhca/dut/DATA.et3"

    # 1. using raw interface:
    #    averaged transfer impedance from raw data
    # et3_data = load(fstr, verbose=True)
    # ts = load_time(fstr)
    # ati = np.abs(et3_data).sum(axis=1)/192.0

    # 2. using DataFrame interface:
    #    's' is seconds (default)
    #    'T' is minute
    et3 = ET3(file_name, trim=True, verbose=False)
    df = et3.to_df()  # rel_date='2019/01/10'
    dp = et3.to_dp()
    dp["ati"] = np.abs(df).sum(axis=1) / 192.0

    # 3. plot
    dp = dp[2000:]
    fig = plt.figure(figsize=(6, 4))
    ax = fig.add_subplot(111)
    axt = ax.twinx()
    ax.plot(dp.index, dp["tle"])
    ax.plot(dp.index, dp["tre"])
    ax.plot(dp.index, dp["tn"])
    axt.plot(dp.index, dp["ati"])
    ax.grid(True)
    ax.legend(["Left ear", "Right ear", "Nasopharyngeal"])

    hfmt = dates.DateFormatter("%y/%m/%d %H:%M")
    ax.xaxis.set_major_formatter(hfmt)
    fig.autofmt_xdate()



# coding: utf-8
# pylint: disable=invalid-name
"""
util functions for 2D EIT
1. generate stimulation lines/patterns
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np


def eit_scan_lines(ne=16, dist=1):
    """
    generate scan matrix

    Parameters
    ----------
    ne: int
        number of electrodes
    dist: int
        distance between A and B (default=1)

    Returns
    -------
    ex_mat: NDArray
        stimulation matrix

    Notes
    -----
    in the scan of EIT (or stimulation matrix), we use 4-electrodes
    mode, where A, B are used as positive and negative stimulation
    electrodes and M, N are used as voltage measurements

    1 (A) for positive current injection,
    -1 (B) for negative current sink

    dist is the distance (number of electrodes) of A to B
    in 'adjacent' mode, dist=1, in 'apposition' mode, dist=ne/2

    Examples
    --------
    # let the number of electrodes, ne=16

    if mode=='neighbore':
        ex_mat = eit_scan_lines()
    elif mode=='apposition':
        ex_mat = eit_scan_lines(dist=8)

    WARNING
    -------
    ex_mat is a local index, where it is ranged from 0...15, within the range
    of the number of electrodes. In FEM applications, you should convert ex_mat
    to global index using the (global) el_pos parameters.
    """
    ex = np.array([[i, np.mod(i + dist, ne)] for i in range(ne)])

    return ex


if __name__ == "__main__":
    m = eit_scan_lines(dist=8)
    print(m)



# pylint: disable=invalid-name, too-many-locals
"""static features of R values measured using EIT system"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

import numpy as np


def ati(x):
    """
    averaged total impedance, unit (mV),
    original / 192.0
    [experimental] new / 16.0

    Notes
    -----
    if I=1mA, then ati returns Ohms
    """
    # implement old behavior of numpy.nansum
    if np.isnan(x).any():
        v = np.nan
    else:
        v = np.sum(np.abs(x)) / 192.0

    return v


def ati_df(x):
    """ati of DataFrame"""
    return x.abs().sum(skipna=False) / 192.0


def fmmu_index(n_el=16, dist=8, step=1):
    """
    generate FMMU patterns and left, right index

    n_el = 16
    dist = 8 # opposition
    step = 1 # adjacent measures
    """
    # [0, ..., 15]
    left_el = [13, 14, 15, 0, 1, 2, 3, 4]
    right_el = [5, 6, 7, 8, 9, 10, 11, 12]

    m_array = []
    for a in range(n_el):
        b = (a + dist) % n_el
        for j in range(a, a + n_el):
            m = j % n_el
            n = (j + step) % n_el
            if not (m in [a, b] or n in [a, b]):
                diff_pair = [n, m]  # v_n - v_m
                m_array.append(diff_pair)

    m_array = np.array(m_array)
    N = m_array.shape[0]
    left_sel = np.zeros(N, dtype=np.bool)  # 192
    right_sel = np.zeros(N, dtype=np.bool)
    for i, nm in enumerate(m_array):
        n = nm[0]
        if n in left_el:
            left_sel[i] = True
        if n in right_el:
            right_sel[i] = True

    return left_sel, right_sel


def ati_lr(x, sel):
    """extract ATI left, right"""
    x_sel = np.nanmean(np.abs(x[sel]))

    return x_sel


def rchannel(x, offset=0):
    """
    calculate R (voltages) of two electrodes counted from the
    excitation electrode with a fixed number (offset)
    """
    N = 12  # measurements within a excitation pattern
    c = np.mod(offset, N)
    x = x[:, c:N:]
    return np.abs(x).sum(axis=1) / float(N)



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-locals
# pylint: disable=too-many-instance-attributes
""" 2D/3D FEM routines """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

from collections import namedtuple
import numpy as np
import numpy.linalg as la
from scipy import sparse

from .utils import eit_scan_lines


class Forward:
    """FEM forward computing code"""

    def __init__(self, mesh, el_pos):
        """
        A good FEM forward solver should only depend on
        mesh structure and the position of electrodes

        Parameters
        ----------
        mesh: dict
            mesh structure, {'node', 'element', 'perm'}
        el_pos: NDArray
            numbering of electrodes positions

        Note
        ----
        1, The nodes are continuous numbered, the numbering of an element is
        CCW (counter-clock-wise).
        2, The Jacobian and the boundary voltages used the SIGN information,
        for example, V56 = V6 - V5 = -V65. If you are using absolute boundary
        voltages for imaging, you MUST normalize it with the signs of v0
        under each current-injecting pattern.
        """
        self.pts = mesh["node"]
        self.tri = mesh["element"]
        self.tri_perm = mesh["perm"]
        self.el_pos = el_pos

        # reference electrodes [ref node should not be on electrodes]
        ref_el = 0
        while ref_el in self.el_pos:
            ref_el = ref_el + 1
        self.ref = ref_el

        # infer dimensions from mesh
        self.n_pts, self.n_dim = self.pts.shape
        self.n_tri, self.n_vertices = self.tri.shape
        self.ne = el_pos.size

    def solve_eit(self, ex_mat=None, step=1, perm=None, parser=None):
        """
        EIT simulation, generate perturbation matrix and forward v

        Parameters
        ----------
        ex_mat: NDArray
            numLines x n_el array, stimulation matrix
        step: int
            the configuration of measurement electrodes (default: adjacent)
        perm: NDArray
            Mx1 array, initial x0. must be the same size with self.tri_perm
        parser: str
            see voltage_meter for more details.

        Returns
        -------
        jac: NDArray
            number of measures x n_E complex array, the Jacobian
        v: NDArray
            number of measures x 1 array, simulated boundary measures
        b_matrix: NDArray
            back-projection mappings (smear matrix)
        """
        # initialize/extract the scan lines (default: apposition)
        if ex_mat is None:
            ex_mat = eit_scan_lines(16, 8)

        # initialize the permittivity on element
        if perm is None:
            perm0 = self.tri_perm
        elif np.isscalar(perm):
            perm0 = np.ones(self.n_tri, dtype=np.float)
        else:
            assert perm.shape == (self.n_tri,)
            perm0 = perm

        # calculate f and Jacobian iteratively over all stimulation lines
        jac, v, b_matrix = [], [], []
        n_lines = ex_mat.shape[0]

        for i in range(n_lines):
            # FEM solver of one stimulation pattern, a row in ex_mat
            ex_line = ex_mat[i]
            f, jac_i = self.solve(ex_line, perm0)
            f_el = f[self.el_pos]

            # boundary measurements, subtract_row-voltages on electrodes
            diff_op = voltage_meter(ex_line, n_el=self.ne, step=step, parser=parser)
            v_diff = subtract_row(f_el, diff_op)
            jac_diff = subtract_row(jac_i, diff_op)

            # build bp projection matrix
            # 1. we can either smear at the center of elements, using
            #    >> fe = np.mean(f[self.tri], axis=1)
            # 2. or, simply smear at the nodes using f
            b = smear(f, f_el, diff_op)

            # append
            v.append(v_diff)
            jac.append(jac_diff)
            b_matrix.append(b)

        # update output, now you can call p.jac, p.v, p.b_matrix
        pde_result = namedtuple("pde_result", ["jac", "v", "b_matrix"])
        p = pde_result(jac=np.vstack(jac), v=np.hstack(v), b_matrix=np.vstack(b_matrix))
        return p

    def solve(self, ex_line, perm):
        """
        with one pos (A), neg(B) driven pairs, calculate and
        compute the potential distribution (complex-valued)

        The calculation of Jacobian can be skipped.
        Currently, only simple electrode model is supported,
        CEM (complete electrode model) is under development.

        Parameters
        ----------
        ex_line: NDArray
            stimulation (scan) patterns/lines
        perm: NDArray
            permittivity on elements (initial)

        Returns
        -------
        f: NDArray
            potential on nodes
        J: NDArray
            Jacobian
        """
        # 1. calculate local stiffness matrix (on each element)
        ke = calculate_ke(self.pts, self.tri)

        # 2. assemble to global K
        kg = assemble_sparse(ke, self.tri, perm, self.n_pts, ref=self.ref)

        # 3. calculate electrode impedance matrix R = K^{-1}
        r_matrix = la.inv(kg)
        r_el = r_matrix[self.el_pos]

        # 4. solving nodes potential using boundary conditions
        b = self._natural_boundary(ex_line)
        f = np.dot(r_matrix, b).ravel()

        # 5. build Jacobian matrix column wise (element wise)
        #    Je = Re*Ke*Ve = (nex3) * (3x3) * (3x1)
        jac = np.zeros((self.ne, self.n_tri), dtype=perm.dtype)
        for (i, e) in enumerate(self.tri):
            jac[:, i] = np.dot(np.dot(r_el[:, e], ke[i]), f[e])

        return f, jac

    def _natural_boundary(self, ex_line):
        """
        Notes
        -----
        Generate the Neumann boundary condition. In utils.py,
        you should note that ex_line is local indexed from 0...15,
        which need to be converted to global node number using el_pos.
        """
        drv_a_global = self.el_pos[ex_line[0]]
        drv_b_global = self.el_pos[ex_line[1]]

        # global boundary condition
        b = np.zeros((self.n_pts, 1))
        b[drv_a_global] = 1.0
        b[drv_b_global] = -1.0

        return b


def smear(f, fb, pairs):
    """
    build smear matrix B for bp

    Parameters
    ----------
    f: NDArray
        potential on nodes
    fb: NDArray
        potential on adjacent electrodes
    pairs: NDArray
        electrodes numbering pairs

    Returns
    -------
    B: NDArray
        back-projection matrix
    """
    b_matrix = []
    for i, j in pairs:
        f_min, f_max = min(fb[i], fb[j]), max(fb[i], fb[j])
        b_matrix.append((f_min < f) & (f <= f_max))

    return np.array(b_matrix)


def subtract_row(v, pairs):
    """
    v_diff[k] = v[i, :] - v[j, :]

    Parameters
    ----------
    v: NDArray
        Nx1 boundary measurements vector or NxM matrix
    pairs: NDArray
        Nx2 subtract_row pairs

    Returns
    -------
    v_diff: NDArray
        difference measurements
    """
    i = pairs[:, 0]
    j = pairs[:, 1]
    # row-wise/element-wise operation on matrix/vector v
    v_diff = v[i] - v[j]

    return v_diff


def voltage_meter(ex_line, n_el=16, step=1, parser=None):
    """
    extract subtract_row-voltage measurements on boundary electrodes.
    we direct operate on measurements or Jacobian on electrodes,
    so, we can use LOCAL index in this module, do not require el_pos.

    Notes
    -----
    ABMN Model.
    A: current driving electrode,
    B: current sink,
    M, N: boundary electrodes, where v_diff = v_n - v_m.

    'no_meas_current': (EIDORS3D)
    mesurements on current carrying electrodes are discarded.

    Parameters
    ----------
    ex_line: NDArray
        2x1 array, [positive electrode, negative electrode].
    n_el: int
        number of total electrodes.
    step: int
        measurement method (two adjacent electrodes are used for measuring).
    parser: str
        if parser is 'fmmu', or 'rotate_meas' then data are trimmed,
        boundary voltage measurements are re-indexed and rotated,
        start from the positive stimulus electrodestart index 'A'.
        if parser is 'std', or 'no_rotate_meas' then data are trimmed,
        the start index (i) of boundary voltage measurements is always 0.

    Returns
    -------
    v: NDArray
        (N-1)*2 arrays of subtract_row pairs
    """
    # local node
    drv_a = ex_line[0]
    drv_b = ex_line[1]
    i0 = drv_a if parser in ("fmmu", "rotate_meas") else 0

    # build differential pairs
    v = []
    for a in range(i0, i0 + n_el):
        m = a % n_el
        n = (m + step) % n_el
        # if any of the electrodes is the stimulation electrodes
        if not (m == drv_a or m == drv_b or n == drv_a or n == drv_b):
            # the order of m, n matters
            v.append([n, m])

    diff_pairs = np.array(v)
    return diff_pairs


def assemble(ke, tri, perm, n_pts, ref=0):
    """
    Assemble the stiffness matrix (dense matrix, default)

    Parameters
    ----------
    ke: NDArray
        n_tri x (n_dim x n_dim) 3d matrix
    tri: NDArray
        the structure of mesh
    perm: NDArray
        n_tri x 1 conductivities on elements
    n_pts: int
        number of nodes
    ref: int
        reference electrode

    Returns
    -------
    K: NDArray
        k_matrix, NxN array of complex stiffness matrix

    Notes
    -----
    you can use sparse matrix (IJV) format to automatically add the local
    stiffness matrix to the global matrix.
    """
    n_tri = tri.shape[0]

    # assemble global stiffness matrix
    k_global = np.zeros((n_pts, n_pts), dtype=perm.dtype)
    for ei in range(n_tri):
        k_local = ke[ei]
        pe = perm[ei]

        no = tri[ei, :]
        ij = np.ix_(no, no)
        k_global[ij] += k_local * pe

    # place reference electrode
    if 0 <= ref < n_pts:
        k_global[ref, :] = 0.0
        k_global[:, ref] = 0.0
        k_global[ref, ref] = 1.0

    return k_global


def assemble_sparse(ke, tri, perm, n_pts, ref=0):
    """
    Assemble the stiffness matrix (using sparse matrix)

    Parameters
    ----------
    ke: NDArray
        n_tri x (n_dim x n_dim) 3d matrix
    tri: NDArray
        the structure of mesh
    perm: NDArray
        n_tri x 1 conductivities on elements
    n_pts: int
        number of nodes
    ref: int
        reference electrode

    Returns
    -------
    K: NDArray
        k_matrix, NxN array of complex stiffness matrix

    Notes
    -----
    you may use sparse matrix (IJV) format to automatically add the local
    stiffness matrix to the global matrix.
    """
    n_tri, n_vertices = tri.shape

    # New: use IJV indexed sparse matrix to assemble K (fast, prefer)
    # index = np.array([np.meshgrid(no, no, indexing='ij') for no in tri])
    # note: meshgrid is slow, using handcraft sparse index, for example
    # let tri=[[1, 2, 3], [4, 5, 6]], then indexing='ij' is equivalent to
    # row = [1, 1, 1, 2, 2, 2, ...]
    # col = [1, 2, 3, 1, 2, 3, ...]
    row = np.repeat(tri, n_vertices).ravel()
    col = np.repeat(tri, n_vertices, axis=0).ravel()
    data = np.array([ke[i] * perm[i] for i in range(n_tri)]).ravel()

    # set reference nodes before constructing sparse matrix, where
    # K[ref, :] = 0, K[:, ref] = 0, K[ref, ref] = 1.
    # write your own mask code to set the corresponding locations of data
    # before building the sparse matrix, for example,
    # data = mask_ref_node(data, row, col, ref)

    # for efficient sparse inverse (csc)
    A = sparse.csr_matrix((data, (row, col)), shape=(n_pts, n_pts), dtype=perm.dtype)

    # the stiffness matrix may not be sparse
    A = A.toarray()

    # place reference electrode
    if 0 <= ref < n_pts:
        A[ref, :] = 0.0
        A[:, ref] = 0.0
        A[ref, ref] = 1.0

    return A


def calculate_ke(pts, tri):
    """
    Calculate local stiffness matrix on all elements.

    Parameters
    ----------
    pts: NDArray
        Nx2 (x,y) or Nx3 (x,y,z) coordinates of points
    tri: NDArray
        Mx3 (triangle) or Mx4 (tetrahedron) connectivity of elements

    Returns
    -------
    ke_array: NDArray
        n_tri x (n_dim x n_dim) 3d matrix
    """
    n_tri, n_vertices = tri.shape

    # check dimension
    # '3' : triangles
    # '4' : tetrahedrons
    if n_vertices == 3:
        _k_local = _k_triangle
    elif n_vertices == 4:
        _k_local = _k_tetrahedron
    else:
        raise TypeError("The num of vertices of elements must be 3 or 4")

    # default data types for ke
    ke_array = np.zeros((n_tri, n_vertices, n_vertices))
    for ei in range(n_tri):
        no = tri[ei, :]
        xy = pts[no]

        # compute the KIJ (permittivity=1.)
        ke = _k_local(xy)
        ke_array[ei] = ke

    return ke_array


def _k_triangle(xy):
    """
    given a point-matrix of an element, solving for Kij analytically
    using barycentric coordinates (simplex coordinates)

    Parameters
    ----------
    xy: NDArray
        (x,y) of nodes 1,2,3 given in counterclockwise manner

    Returns
    -------
    ke_matrix: NDArray
        local stiffness matrix
    """
    # edges (vector) of triangles
    s = xy[[2, 0, 1]] - xy[[1, 2, 0]]
    # s1 = xy[2, :] - xy[1, :]
    # s2 = xy[0, :] - xy[2, :]
    # s3 = xy[1, :] - xy[0, :]

    # area of triangles. Note, abs is removed since version 2020,
    # user must make sure all triangles are CCW (conter clock wised).
    # at = 0.5 * la.det(s[[0, 1]])
    at = 0.5 * det2x2(s[0], s[1])

    # (e for element) local stiffness matrix
    ke_matrix = np.dot(s, s.T) / (4.0 * at)

    return ke_matrix


def det2x2(s1, s2):
    """Calculate the determinant of a 2x2 matrix"""
    return s1[0] * s2[1] - s1[1] * s2[0]


def _k_tetrahedron(xy):
    """
    given a point-matrix of an element, solving for Kij analytically
    using barycentric coordinates (simplex coordinates)

    Parameters
    ----------
    xy: NDArray
        (x,y) of nodes 1, 2, 3, 4 given in counterclockwise manner,
        see notes.

    Returns
    -------
    ke_matrix: NDArray
        local stiffness matrix

    Notes
    -----
    A tetrahedron is described using [0, 1, 2, 3] (local node index) or
    [171, 27, 9, 53] (global index). Counterclockwise (CCW) is defined
    such that the barycentric coordinate of face (1->2->3) is positive.
    """
    s = xy[[2, 3, 0, 1]] - xy[[1, 2, 3, 0]]

    # volume of the tetrahedron, Note abs is removed since version 2020,
    # user must make sure all tetrahedrons are CCW (counter clock wised).
    vt = 1.0 / 6 * la.det(s[[0, 1, 2]])

    # calculate area (vector) of triangle faces
    # re-normalize using alternative (+,-) signs
    ij_pairs = [[0, 1], [1, 2], [2, 3], [3, 0]]
    signs = [1, -1, 1, -1]
    a = [sign * np.cross(s[i], s[j]) for (i, j), sign in zip(ij_pairs, signs)]
    a = np.array(a)

    # local (e for element) stiffness matrix
    ke_matrix = np.dot(a, a.transpose()) / (36.0 * vt)

    return ke_matrix



# pylint: disable=no-member, invalid-name
"""
open/view .mes file (a binary mesh structure)

The mesh structure in mes was developed by FMMU EIT group (Bin Yang, et al.)
Please Cite the following paper if you are using .mes in your research:
    Yang, Bin, et al. "Comparison of electrical impedance tomography and
    intracranial pressure during dehydration treatment of cerebral edema."
    NeuroImage: Clinical 23 (2019): 101909.
"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
import ctypes
import struct
import numpy as np
import matplotlib.pyplot as plt
from pkg_resources import resource_filename


def load(fstr, mirror=False):
    """
    Parameters
    ----------
    mirror: mirror left and right electrodes

    return mesh and el_pos
    """

    # note: seek is global position,
    # fh can be advanced using read in sub-function
    with open(fstr, "rb") as fh:

        # 0. extract BMP
        # (offset=bmp_size)
        bmp_size = get_bmp_size(fh)
        bmp = bytearray(fh.read(bmp_size))
        save_bmp(fstr, bmp)

        # 1. extract mesh [104 Bytes / structure]
        # ne = tri.shape[0]
        # (offset = 4 + ne*104)
        tri, perm = extract_element(fh)

        # 2. extract nodes [20 Bytes / structure]
        # nn = pts.shape[0]
        # (offset=4 + nn*20)
        pts = extract_node(fh)

        # 3. extract electrodes
        el_pos = extract_el(fh)

    if mirror:
        ne = np.size(el_pos)
        ne_start = np.int(ne / 2)
        el_index = np.mod(np.arange(ne_start, -ne_start, -1), ne)
        el_pos = el_pos[el_index]

    mesh = {"node": pts, "element": tri, "perm": perm}
    return mesh, el_pos


def get_bmp_size(fh):
    """get the size of bmp segment"""
    # size of BMP is stored in a UNSIGNED LONG LONG at the end of .mes
    nff = ctypes.sizeof(ctypes.c_int64)
    # seek backwards from the end (2) of the file
    fh.seek(-nff, 2)
    # unsigned long long is 'Q'
    bmp_size = struct.unpack("Q", fh.read(nff))[0]
    # fh is a file ID,
    # it can be modified via read within subroutine, rewind!
    fh.seek(0)

    return bmp_size


def save_bmp(fstr, bmp):
    """save bmp segment to file"""
    # automatically infer file name from 'fstr'
    bmp_file = fstr.replace(".mes", ".bmp")
    with open(bmp_file, "wb") as fh:
        fh.write(bmp)


def extract_element(fh):
    """
    extract element segment

    Notes
    -----
    structure of each element:
    {
        int node1;       // node-1 (global numbering)
        int node2;       // node-2 (global numbering)
        int node3;       // node-3 (global numbering)
        int gENum;       // element (global numbering)
        double alpha;    // conductive (sigma)
        double Ae[3][3]; // sensitivity matrix (local) see. Tadakuni murai
        double deltae;   // area of this element
    }
    = 104 Bytes
    """
    # number of element
    nff = ctypes.sizeof(ctypes.c_int)
    ne = struct.unpack("i", fh.read(nff))[0]
    tri = np.zeros((ne, 3), dtype="int")
    perm = np.zeros(ne, dtype="double")

    for _ in range(ne):
        d = np.array(struct.unpack("4i10dd", fh.read(104)))
        # global element number
        ge_num = int(d[3])
        tri[ge_num, :] = d[:3]
        perm[ge_num] = d[4]

    return tri, perm


def extract_node(fh):
    """
    extract node structure

    Notes
    -----
    structure of each node
    {
        double x; // x-coordinate
        double y; // y-coordinate
        int gNum; // node (global numbering)
    }
    = 20 Bytes
    """
    # number of nodes
    nff = ctypes.sizeof(ctypes.c_int)
    nn = struct.unpack("i", fh.read(nff))[0]
    pts = np.zeros((nn, 2), dtype="double")

    for _ in range(nn):
        d = np.array(struct.unpack("2di", fh.read(20)))
        # global node number
        gn_num = int(d[2])
        # offset to overlay on .bmp file
        # x_new = x+8, y_new = -y-8 (yang bin)
        # x_new = x-8, y_new = -y-8 (lby)
        pts[gn_num, :] = [d[0] - 4.0, -d[1] - 4.0]

    return pts


def extract_el(fh):
    """extract the locations of electrodes"""
    # how many electrodes in .mes
    nff = ctypes.sizeof(ctypes.c_int)
    ne = struct.unpack("i", fh.read(nff))[0]
    # read all at once
    el_pos = np.array(struct.unpack(ne * "i", fh.read(ne * 4)))

    return el_pos


if __name__ == "__main__":
    # How to load and use a .mes file (github.com/liubenyuan/pyeit-models)
    mstr = resource_filename("pyeit-models", "data/model/DLS2.mes")
    mesh_obj, el_pos = load(fstr=mstr)

    # print the size
    e, pts = mesh_obj["element"], mesh_obj["node"]
    # print('tri size = (%d, %d)' % e.shape)
    # print('pts size = (%d, %d)' % pts.shape)

    # show mesh
    fig, ax = plt.subplots(1, figsize=(6, 6))
    ax.triplot(pts[:, 0], pts[:, 1], e)
    ax.plot(pts[el_pos, 0], pts[el_pos, 1], "ro")
    for i, el in enumerate(el_pos):
        ax.text(pts[el, 0], pts[el, 1], str(i + 1), color="r")
    ax.set_aspect("equal")
    ax.invert_yaxis()

    # bmp and mesh overlay
    fig, ax = plt.subplots(figsize=(6, 6))
    imstr = mstr.replace("mes", "bmp")
    im = plt.imread(imstr)
    ax.imshow(im)
    ax.set_aspect("equal")

    # the plot will automatically align with an overlay image
    ax.triplot(pts[:, 0], pts[:, 1], e)
    ax.plot(pts[el_pos, 0], pts[el_pos, 1], "ro")
    for i, el in enumerate(el_pos):
        ax.text(pts[el, 0], pts[el, 1], str(i + 1), color="r")
    ax.axis("off")
    plt.show()



# pylint: disable=no-member, invalid-name
# pylint: disable=too-many-arguments, too-many-locals
"""plot EIT data"""
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function
import os.path

import numpy as np
import matplotlib.pyplot as plt
from matplotlib import dates


def mesh_plot(
    mesh,
    el_pos,
    mstr="",
    figsize=(9, 6),
    alpha=0.5,
    offset_ratio=0.075,
    show_image=False,
    show_mesh=False,
    show_electrode=True,
    show_number=False,
    show_text=True,
):
    """plot mesh structure (base layout)"""
    # load mesh structure
    pts = mesh["node"]
    tri = mesh["element"]
    fig, ax = plt.subplots(figsize=figsize)
    ax.set_facecolor("black")
    ax.set_aspect("equal")

    # load background
    if show_image and os.path.exists(mstr):
        image_name = mstr.replace("mes", "bmp")
        im = plt.imread(image_name)
        ax.imshow(im)
    else:
        # without imshow, the yaxis should be inverted
        ax.invert_yaxis()

    # show mesh structure
    if show_mesh:
        ax.triplot(pts[:, 0], pts[:, 1], tri, alpha=alpha)

    # show electrodes markers
    if show_electrode:
        ax.plot(pts[el_pos, 0], pts[el_pos, 1], "yo")

    # annotate electrodes numbering
    if show_number:
        for i, e in enumerate(el_pos):
            ax.text(pts[e, 0], pts[e, 1], np.str(i + 1), color="r", size=12)

    # annotate (L) at offset_ratio*d beside node 0
    if show_text:
        xb, xa = pts[el_pos[8], 0], pts[el_pos[0], 0]
        d = np.abs(xa - xb)
        offset = d * offset_ratio
        x, y = xa + offset, pts[el_pos[0], 1]
        ax.text(x, y, "L", size=20, color="w")
        # enlarge the right of axes if using annotation
        ax.set_xlim([xb - offset, xa + 2 * offset])

    # clean up axis
    ax.grid("off")
    plt.setp(ax.get_xticklines(), visible=False)
    plt.setp(ax.get_xticklabels(), visible=False)
    plt.setp(ax.get_yticklines(), visible=False)
    plt.setp(ax.get_yticklabels(), visible=False)

    return fig, ax


def ts_plot(ts, figsize=(6, 4), ylabel="ATI (Ohm)", ylim=None, xdate_format=True):
    """plot time series data"""
    fig, ax = plt.subplots(figsize=figsize)
    ax.plot(ts)
    ax.grid("on")
    ax.set_ylabel(ylabel)
    if ylim is not None:
        assert len(ylim) == 2
        ax.set_ylim(ylim)
    # for better xticklabels
    if xdate_format:
        axis_format = dates.DateFormatter("%m/%d %H:%M")
        ax.xaxis.set_major_formatter(axis_format)
        fig.autofmt_xdate()

    return fig, ax



# coding: utf-8
# pylint: disable=invalid-name, no-member, arguments-differ
""" bp (back-projection) and f(filtered)-bp module """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np
from .base import EitBase


class BP(EitBase):
    """A naive inversion of (Euclidean) back projection."""

    def setup(self, weight="none"):
        """setup BP"""
        self.params = {"weight": weight}

        # build the weighting matrix
        # BP: in node imaging, H is the smear matrix (transpose of B)
        if weight == "simple":
            weights = self.simple_weight(self.B.shape[0])
            wb_mat = weights * self.B
            self.H = wb_mat.T

    def normalize(self, v1, v0):
        """redefine normalize for BP (without amplitude normalization)"""
        dv = (v1 - v0) / self.v0_sign
        return dv

    def map(self, dv):
        """return Hx"""
        x = -dv / self.v0_sign
        return np.dot(self.H, x.transpose())

    def solve_gs(self, v1, v0):
        """solving using gram-schmidt"""
        a = np.dot(v1, v0) / np.dot(v0, v0)
        vn = -(v1 - a * v0) / self.v0_sign
        ds = np.dot(self.H, vn.transpose())
        return ds

    def simple_weight(self, num_voltages):
        """
        building weighting matrix : simple, normalize by radius.

        Note
        ----
        as in fem.py, we could either smear at,

        (1) elements, using the center co-ordinates (x,y) of each element
            >> center_e = np.mean(self.pts[self.tri], axis=1)
        (2) nodes.

        Parameters
        ----------
        num_voltages: int
            number of equal-potential lines

        Returns
        -------
        w: NDArray
            weighting matrix
        """
        d = np.sqrt(np.sum(self.pts**2, axis=1))
        r = np.max(d)
        w = (1.01 * r - d) / (1.01 * r)
        # weighting by element-wise multiplication W with B
        weights = np.dot(np.ones((num_voltages, 1)), w.reshape(1, -1))
        return weights



"""
The :mod:`pyeit.feature_extraction` module deals with feature extraction
from raw EIT data. It currently includes methods to extract features from
EIT dynamic images and static properties.
"""

from .transfer_impedance import ati, fmmu_index, ati_lr, ati_df
from .mesh_geometry import SimpleMeshGeometry

__all__ = ["ati", "fmmu_index", "ati_lr", "ati_df", "SimpleMeshGeometry"]



""" main module for 2D/3D mesh """
from .wrapper import create, set_perm, layer_circle
from .shell import multi_shell, multi_circle

__all__ = ["create", "set_perm", "layer_circle", "multi_shell", "multi_circle"]



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-locals
""" implement distance functions for distmesh """
# Copyright (c) Benyuan Liu. All rights reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np

from .utils import dist, edge_project
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon


def circle(pts, pc=None, r=1.0):
    """
    Distance function for the circle centered at pc = [xc, yc]

    Parameters
    ----------
    pts : array_like
        points on 2D
    pc : array_like, optional
        center of points
    r : float, optional
        radius

    Returns
    -------
    array_like
        distance of (points - pc) - r

    Note
    ----
    copied and modified from https://github.com/ckhroulev/py_distmesh2d
    """
    if pc is None:
        pc = [0, 0]
    return dist(pts - pc) - r


def ellipse(pts, pc=None, ab=None):
    """Distance function for the ellipse
    centered at pc = [xc, yc], with a, b = [a, b]
    """
    if pc is None:
        pc = [0, 0]
    if ab is None:
        ab = [1.0, 2.0]
    return dist((pts - pc) / ab) - 1.0


def unit_circle(pts):
    """unit circle at (0, 0)"""
    return circle(pts, r=1.0)


def box_circle(pts):
    """unit circle at (0.5, 0.5) with r=0.5"""
    return circle(pts, pc=[0.5, 0.5], r=0.5)


def ball(pts, pc=None, r=1.0):
    """
    generate balls in 3D (default: unit ball)

    See Also
    --------
    circle : generate circles in 2D
    """
    if pc is None:
        pc = [0, 0, 0]
    return circle(pts, pc, r)


def unit_ball(pts):
    """generate unit ball in 3D"""
    return ball(pts)


def rectangle0(pts, p1=None, p2=None):
    """
    Distance function for the rectangle p1=[x1, y1] and p2=[x2, y2]

    Note
    ----
    p1 should be bottom-left, p2 should be top-right
    if p in rect(p1, p2), then (p-p1)_x and (p-p2)_x must have opposite sign

    Parameters
    ----------
    pts : array_like
    p1 : array_like, optional
        bottom left coordinates
    p2 : array_like, optional
        top tight coordinates

    Returns
    -------
    array_like
        distance
    """
    if p1 is None:
        p1 = [0, 0]
    if p2 is None:
        p2 = [1, 1]
    if pts.ndim == 1:
        pts = pts[np.newaxis]
    pd_left = [-min(row) for row in pts - p1]
    pd_right = [max(row) for row in pts - p2]

    return np.maximum(pd_left, pd_right)


def rectangle(pts, p1=None, p2=None):
    """
    smoothed rectangle

    p1: buttom-left corner
    p2: top-right corner
    """
    if p1 is None:
        p1 = [0, 0]
    if p2 is None:
        p2 = [1, 1]
    if pts.ndim == 1:
        pts = pts[np.newaxis]

    d1x = -(pts[:, 0] - p1[0])
    d2x = pts[:, 0] - p2[0]
    d1y = -(pts[:, 1] - p1[1])
    d2y = pts[:, 1] - p2[1]

    # find interior points (d < 0)
    dx = np.maximum(d1x, d2x)
    dy = np.maximum(d1y, d2y)
    d = np.maximum(dx, dy)

    # smoothed corner distance function
    ix_left = d1x > 0
    ix_right = d2x > 0
    iy_below = d1y > 0
    iy_above = d2y > 0

    # 1, 2, 3, 4 Quadrant outside-points smooth distance
    ix1 = np.logical_and(ix_right, iy_above)
    d[ix1] = np.sqrt(d2x[ix1] ** 2 + d2y[ix1] ** 2)
    ix2 = np.logical_and(ix_left, iy_above)
    d[ix2] = np.sqrt(d1x[ix2] ** 2 + d2y[ix2] ** 2)
    ix3 = np.logical_and(ix_left, iy_below)
    d[ix3] = np.sqrt(d1x[ix3] ** 2 + d1y[ix3] ** 2)
    ix4 = np.logical_and(ix_right, iy_below)
    d[ix4] = np.sqrt(d2x[ix4] ** 2 + d1y[ix4] ** 2)

    return d


def fix_points_fd(fd, n_el=16, pc=None):
    """
    return fixed and uniformly distributed points on
    fd with equally distributed angles

    Parameters
    ----------
    fd : distance function
    pc : array_like, optional
        center of points
    n_el : number of electrodes, optional

    Returns
    -------
    array_like
        coordinates of fixed points
    """
    if pc is None:
        pc = [0, 0]

    # initialize points
    r = 10.0
    theta = 2.0 * np.pi * np.arange(n_el) / float(n_el)
    # add offset of theta
    # theta += theta[1] / 2.0
    p_fix = [[-r * np.cos(th), r * np.sin(th)] for th in theta]
    pts = np.array(p_fix) + pc

    # project back on edges
    pts_new = np.inf * np.ones_like(pts)
    c = False
    d_eps = 0.1
    max_iter = 10
    niter = 0
    while not c:
        # project on fd
        pts_new = edge_project(pts, fd)
        # project on rays
        r = dist(pts_new)
        pts_new = [[-ri * np.cos(ti), ri * np.sin(ti)] for ri, ti in zip(r, theta)]
        pts_new = np.array(pts_new)
        # check convergence
        c = np.sum(dist(pts_new - pts)) < d_eps or niter > max_iter
        pts = pts_new
        niter += 1
    return pts_new


def fix_points_circle(pc=None, offset=0, r=1.0, ppl=16):
    """
    return fixed and uniformly distributed points on
    a circle with radius r

    Parameters
    ----------
    pc : array_like, optional
        center of points
    r : float, optional
        radius
    ppl : number of points, optional

    Returns
    -------
    array_like
        coordinates of fixed points
    """
    if pc is None:
        pc = [0, 0]

    delta_theta = 2.0 * np.pi / float(ppl)
    theta = np.arange(ppl) * delta_theta + delta_theta * offset
    p_fix = [[-r * np.cos(th), r * np.sin(th)] for th in theta]
    return np.array(p_fix) + pc


def fix_points_ball(pc=None, r=1.0, z=0.0, n_el=16):
    """
    return fixed and uniformly distributed points on
    a circle with radius r

    Parameters
    ----------
    pc : array_like,
        center of points
    r : float,
        radius
    z : float,
        z level of points
    n_el : number of electrodes, optional

    Returns
    -------
    array_like
        coordinates of fixed points
    """
    if pc is None:
        pc = [0, 0, 0]

    ry = np.sqrt(r**2 - z**2)
    theta = 2.0 * np.pi * np.arange(n_el) / float(n_el)
    p_fix = [[ry * np.sin(th), ry * np.cos(th), z] for th in theta]
    return np.array(p_fix) + pc


def dist_diff(d1, d2):
    """Distance function for the difference of two sets.

    Parameters
    ----------
    d1 : array_like
    d2 : array_like
        distance of two functions

    Returns
    -------
    array_like
        maximum difference

    Note
    ----
    boundary is denoted by d=0
    copied and modified from https://github.com/ckhroulev/py_distmesh2d
    """
    return np.maximum(d1, -d2)


def dist_intersect(d1, d2):
    """Distance function for the intersection of two sets.

    Parameters
    ----------
    d1 : array_like
    d2 : array_like
        distance of two functions

    Returns
    -------
    array_like

    Note
    ----
    boundary is denoted by d=0
    copied and modified from https://github.com/ckhroulev/py_distmesh2d
    """
    return np.maximum(d1, d2)


def dist_union(d1, d2):
    """Distance function for the union of two sets.

    Parameters
    ----------
    d1 : array_like
    d2 : array_like
        distance of two functions

    Returns
    -------
    array_like

    Note
    ----
    boundary is denoted by d=0
    copied and modified from https://github.com/ckhroulev/py_distmesh2d
    """
    return np.minimum(d1, d2)


def area_uniform(p):
    """uniform mesh distribution

    Parameters
    ----------
    p : array_like
        points coordinates

    Returns
    -------
    array_like
        ones

    """
    return np.ones(p.shape[0])


def thorax(pts):

    """
     p : Nx2 ndarray

    returns boolean ndarray specifiying whether the point is inside thorax or not
    e.g :
    array([[False, False],
    [ True,  True],
    [ True,  True]])

    """
    # Thorax contour points coordinates are taken from a thorax simulation based on EIDORS
    thrx = Polygon(
        [
            (0.0487, 0.6543),
            (0.1564, 0.6571),
            (0.2636, 0.6697),
            (0.3714, 0.6755),
            (0.479, 0.6686),
            (0.5814, 0.6353),
            (0.6757, 0.5831),
            (0.7582, 0.5137),
            (0.8298, 0.433),
            (0.8894, 0.3431),
            (0.9347, 0.2452),
            (0.9698, 0.1431),
            (0.9938, 0.0379),
            (1.0028, -0.0696),
            (0.9914, -0.1767),
            (0.9637, -0.281),
            (0.9156, -0.3771),
            (0.8359, -0.449),
            (0.7402, -0.499),
            (0.6432, -0.5463),
            (0.5419, -0.5833),
            (0.4371, -0.6094),
            (0.3308, -0.6279),
            (0.2243, -0.6456),
            (0.1168, -0.6508),
            (0.0096, -0.6387),
            (-0.098, -0.6463),
            (-0.2058, -0.6433),
            (-0.313, -0.6312),
            (-0.4181, -0.6074),
            (-0.5164, -0.5629),
            (-0.6166, -0.5232),
            (-0.7207, -0.4946),
            (-0.813, -0.4398),
            (-0.8869, -0.3614),
            (-0.933, -0.2647),
            (-0.9451, -0.1576),
            (-0.9425, -0.0498),
            (-0.9147, 0.0543),
            (-0.8863, 0.1585),
            (-0.8517, 0.2606),
            (-0.8022, 0.3565),
            (-0.7413, 0.4455),
            (-0.6664, 0.5231),
            (-0.5791, 0.5864),
            (-0.4838, 0.6369),
            (-0.3804, 0.667),
            (-0.2732, 0.6799),
            (-0.1653, 0.6819),
            (-0.0581, 0.6699),
        ]
    )

    pts_ = [Point(p[0], p[1]) for p in pts]
    # ba : boolean "array" , will be then converted to an array
    # initialized as list because it's smoother in initialization
    # compared to an array (in first append : axis=1, >=second :axis=0)
    ba = [[thrx.contains(pt), thrx.contains(pt)] for pt in pts_]
    ba = np.array(ba)

    return ba


# L_shaped mesh (for testing)
def L_shaped(pts):
    return dist_diff(
        rectangle(pts, p1=[-1, -1], p2=[1, 1]), rectangle(pts, p1=[0, 0], p2=[1, 1])
    )



""" IO module """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from .et3 import ET3
from .et4 import ET4
from .ewd import EWD
from .daeger_eit import DAEGER_EIT
from .utils import get_date_from_folder

# meshes and auxillary files
from .mes import load as mes_load
from .icp import load as icp_load

__all__ = [
    "ET3",
    "ET4",
    "EWD",
    "DAEGER_EIT",
    "get_date_from_folder",
    "mes_load",
    "icp_load",
]



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-locals, no-name-in-module
""" interpolation on 2D/3D irregular/regular grids """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
from __future__ import division, absolute_import, print_function

import numpy as np
import scipy.linalg as la
from scipy.sparse import coo_matrix
from scipy.spatial import ConvexHull
from matplotlib.path import Path
import matplotlib.pyplot as plt

# for debugging
from pyeit.mesh import layer_circle, set_perm


def meshgrid(pts, n=32, ext_ratio=0, gc=False):
    """
    build xg, yg, mask grids from triangles point cloud
    function for interpolating regular grids

    Parameters
    ----------
    pts: NDArray
        nx2 array of points (x, y)
    el_pos: NDArray (optional)
        the location of electrodes (for extract the convex hull of pts)
    n: int
        the number of meshgrid per dimension
    ext_ratio: float
        extend the boundary of meshgrid by ext_ratio*d
    gc: bool
        grid_correction, offset xgrid and ygrid by half step size

    Notes
    -----
    mask denotes points outside mesh.
    """
    xg, yg = _build_grid(pts, n=n, ext_ratio=ext_ratio, gc=gc)
    # pts_edges = pts[el_pos]
    pts_edges = _hull_points(pts)
    mask = _build_mask(pts_edges, xg, yg)
    return xg, yg, mask


def _build_grid(pts, n=32, ext_ratio=0, gc=False):
    """generating mesh grids"""
    x, y = pts[:, 0], pts[:, 1]
    x_min, x_max = min(x), max(x)
    y_min, y_max = min(y), max(y)
    x_ext = (x_max - x_min) * ext_ratio
    y_ext = (y_max - y_min) * ext_ratio
    xv, xv_step = np.linspace(
        x_min - x_ext, x_max + x_ext, num=n, endpoint=False, retstep=True
    )
    yv, yv_step = np.linspace(
        y_min - y_ext, y_max + y_ext, num=n, endpoint=False, retstep=True
    )
    # if need grid correction
    if gc:
        xv = xv + xv_step / 2.0
        yv = yv + yv_step / 2.0
    xg, yg = np.meshgrid(xv, yv, sparse=False, indexing="xy")
    return xg, yg


def _build_mask(pts_edges, xg, yg):
    """find whether meshgrids is interior of mesh"""
    # 1. create mask based on meshes
    points = np.vstack((xg.flatten(), yg.flatten())).T

    # 2. extract edge points using el_pos
    path = Path(pts_edges, closed=False)
    mask = path.contains_points(points)

    return ~mask


def _hull_points(pts):
    """return the convex hull points"""
    cv = ConvexHull(pts)
    hull_nodes = cv.vertices
    return pts[hull_nodes, :]


def _distance2d(x, y, center="mean"):
    """
    Calculate radius given center.
    This function can be OPTIMIZED using numba or cython.
    """
    if center is None:
        xc, yc = 0, 0
    elif center == "mean":
        xc, yc = np.mean(x), np.mean(y)
    else:
        xc, yc = center[0], center[1]

    d = np.sqrt((x - xc) ** 2 + (y - yc) ** 2).ravel()
    return d


def _distance_matrix2d(xy, xyi):
    """
    Description
    -----------
    (2D only) return element-wise distance matrix (pair-wise)
    """
    # Make a distance matrix between pairwise observations
    # Note: from <http://stackoverflow.com/questions/1871536>
    # (Yay for ufuncs!)
    d0 = np.subtract.outer(xy[:, 0], xyi[:, 0])  # size(xy) * size(xyi)
    d1 = np.subtract.outer(xy[:, 1], xyi[:, 1])

    # hypot : element-wise sqrt(d0**2 + d1**2)
    return np.hypot(d0, d1)


def weight_sigmod(xy, xyi, ratio=0.05, s=20.0):
    """
    Description
    -----------
    (2D only)
    local weight/interpolate by sigmod function (GREIT3D)

    Parameters
    ----------
    xy: NDArray
        (x, y) of values
    xyi: NDArray
        (xi, yi) of interpolated locations
    ratio: float
        R0 = d_max * ratio
    s: float
        control the decay ratio

    Returns
    -------
    w_mat: NDArray
        weighting matrix mapping from xy to xyi (xy meshgrid)
    """
    d_mat = _distance_matrix2d(xy, xyi)
    # normalize distance
    d_max = np.max(d_mat)
    d_mat = 5.0 * d_mat / d_max
    # desired radius (a ratio of max pairwise distance)
    r0 = 5.0 * ratio
    # weights is the sigmod function
    weight = 1.0 / (1 + np.exp(s * (d_mat - r0)))
    # normalized
    w_mat = weight / weight.sum(axis=0)

    return w_mat


def weight_idw(xy, xyi, k=6, p=1.0):
    """
    Description
    -----------
    (2D only)
    local weight/interpolate by inverse distance

    Parameters
    ----------
    xy: NDArray
        (x, y) of values
    xyi: NDArray
        (xi, yi) of interpolated locations
    k: int
        number of nearest neighbores
    p: float
        scaling distance

    Returns
    -------
    w_mat: NDArray
        weighting matrix mapping from xy to xy_mesh
    """
    d_mat = _distance_matrix2d(xy, xyi)
    # weight = 1.0 / d_mat**p
    weight = 1.0 / d_mat**p
    # keep only k largest neighbores (nearest)
    for w in weight.T:
        sort_indices = np.argsort(w)
        np.put(w, sort_indices[:-k], 0)
    # normalized
    w_mat = weight / weight.sum(axis=0)

    # xy times xyi size, use w_mat.T to multiply
    return w_mat


def weight_linear_rbf(xy, xyi, z):
    """
    Description
    -----------
    (2D only)
    local weight/interpolate by linear rbf function (z value required)

    Parameters
    ----------
    xy: NDArray
        (x, y) of values
    xyi: NDArray
        (xi, yi) of interpolated locations

    Returns
    -------
    w_mat: NDArray
        weighting matrix mapping from xy to xy_mesh
    """
    internal_dist = _distance_matrix2d(xy, xy)
    weights = la.solve(internal_dist, z)

    interp_dist = _distance_matrix2d(xy, xyi)
    zi = np.dot(interp_dist.T, weights)

    return zi


def weight_barycentric_gradient():
    """
    Description
    -----------
    (2D only)
    local weight/interpolate by barycentric gradient

    Parameters
    ----------
    xy: NDArray
        (x, y) of values
    xyi: NDArray
        (xi, yi) of interpolated locations

    Returns
    -------
    w_mat: NDArray
        weighting matrix mapping from xy to xy_mesh
    """
    raise NotImplementedError()


def sim2pts(pts, sim, sim_values):
    """
    Description
    -----------
    (2D/3D) compatible.

    Interp values on points using values on simplex,
    a simplex can be triangle or tetrahedron.
    The areas/volumes are used as weights.

    f_n = (sum_e r_e*S_e) / (sum_e S_e)

    where r_e is the value on triangles who share the node n,
    S_e is the area of triangle e.

    Notes
    -----
    This function is similar to pdeprtni of MATLAB pde.
    """
    N = pts.shape[0]
    M, dim = sim.shape
    # calculate the weights
    # triangle/tetrahedron must be CCW (recommended), then a is positive
    if dim == 3:
        weight_func = tri_area
    elif dim == 4:
        weight_func = tet_volume
    weights = weight_func(pts, sim)
    # build tri->pts matrix, could be accelerated using sparse matrix
    row = np.ravel(sim)
    col = np.repeat(np.arange(M), dim)  # [0, 0, 0, 1, 1, 1, ...]
    data = np.repeat(weights, dim)
    e2n_map = coo_matrix((data, (row, col)), shape=(N, M)).tocsr()
    # map values from elements to nodes
    # and re-weight by the sum of the areas/volumes of adjacent elements
    f = e2n_map.dot(sim_values)
    w = np.sum(e2n_map.toarray(), axis=1)

    return f / w


def pts2sim(sim, pts_values):
    """
    Description
    -----------
    (2D/3D) compatible.

    Given values on nodes, calculate interpolated values on simplex,
    this function was tested and equivalent to MATLAB 'pdeintrp'
    except for the shapes of 'pts' and 'tri'

    Parameters
    ----------
    sim: NDArray
        Mx3, Mx4 array, elements or simplex
        triangles denote connectivity [[i, j, k]]
        tetrahedrons denote connectivity [[i, j, m, n]]
    pts_values: NDArray
        Nx1 array, real/complex valued

    Returns
    -------
    el_value: NDArray
        Mx1 array, real/complex valued

    Notes
    -----
    This function is similar to pdfinterp of MATLAB pde.
    """
    # averaged over 3 nodes of a triangle
    el_value = np.mean(pts_values[sim], axis=1)
    return el_value


def tri_area(pts, sim):
    """
    calculate the area of each triangle

    Parameters
    ----------
    pts: NDArray
        Nx2 array, (x,y) locations for points
    sim: NDArray
        Mx3 array, elements (triangles) connectivity

    Returns
    -------
    a: NDArray
        Areas of triangles
    """
    a = np.zeros(np.shape(sim)[0])
    for i, e in enumerate(sim):
        xy = pts[e]
        # s1 = xy[2, :] - xy[1, :]
        # s2 = xy[0, :] - xy[2, :]
        # s3 = xy[1, :] - xy[0, :]
        # which can be simplified to
        # s = xy[[2, 0, 1]] - xy[[1, 2, 0]]
        s = xy[[2, 0]] - xy[[1, 2]]

        # a should be positive if triangles are CCW arranged
        a[i] = la.det(s)

    return a * 0.5


def tet_volume(pts, sim):
    """
    calculate the area of each triangle

    Parameters
    ----------
    pts: NDArray
        Nx3 array, (x,y, z) locations for points
    sim: NDArray
        Mx4 array, elements (tetrahedrons) connectivity

    Returns
    -------
    v: NDArray
        Volumes of tetrahedrons
    """
    v = np.zeros(np.shape(sim)[0])
    for i, e in enumerate(sim):
        xyz = pts[e]
        s = xyz[[2, 3, 0]] - xyz[[1, 2, 3]]

        # a should be positive if triangles are CCW arranged
        v[i] = la.det(s)

    return v / 6.0


def pdetrg(pts, tri):
    """
    Description
    -----------
    (Deprecated)
    analytical calculate the Area and grad(phi_i) using
    barycentric coordinates (simplex coordinates)
    this function is tested and equivalent to MATLAB 'pdetrg'
    except for the shape of 'pts' and 'tri' and the output

    note: each node may have multiple gradients in neighbor
    elements' coordinates. you may averaged all the gradient to
    get one node gradient.

    Parameters
    ----------
    pts: NDArray
        Nx2 array, (x,y) locations for points
    tri: NDArray
        Mx3 array, elements (triangles) connectivity

    Returns
    -------
    a: NDArray
        Mx1 array, areas of elements
    grad_phi_x: NDArray
        Mx3 array, x-gradient on elements' local coordinate
    grad_phi_y: NDArray
        Mx3 array, y-gradient on elements' local coordinate
    """
    m = np.size(tri, 0)

    ix = tri[:, 0]
    iy = tri[:, 1]
    iz = tri[:, 2]

    s1 = pts[iz, :] - pts[iy, :]
    s2 = pts[ix, :] - pts[iz, :]
    s3 = pts[iy, :] - pts[ix, :]

    a = 0.5 * (s2[:, 0] * s3[:, 1] - s3[:, 0] * s2[:, 1])
    if any(a) < 0:
        raise ValueError("Triangles are not in CCW order")

    # note in python, reshape place elements first on the right-most index
    grad_phi_x = np.reshape(
        [-s1[:, 1] / (2.0 * a), -s2[:, 1] / (2.0 * a), -s3[:, 1] / (2.0 * a)], [-1, m]
    ).T
    grad_phi_y = np.reshape(
        [s1[:, 0] / (2.0 * a), s2[:, 0] / (2.0 * a), s3[:, 0] / (2.0 * a)], [-1, m]
    ).T

    return a, grad_phi_x, grad_phi_y


def pdegrad(pts, tri, node_value):
    """
    Description
    -----------
    (Deprecated)
    given values on nodes, calculate the averaged-grad on elements
    this function was tested and equivalent to MATLAB 'pdegrad'
    except for the shape of 'pts', 'tri'

    Parameters
    ----------
    pts: NDArray
        Nx2 array, (x,y) locations for points
    tri: NDArray
        Mx3 array, elements (triangles) connectivity
    node_value: NDArray
        Nx1 array, real/complex valued

    Returns
    -------
    el_grad: NDArray
        el_grad, Mx2 array, real/complex valued
    """
    m = np.size(tri, 0)
    _, grad_phi_x, grad_phi_y = pdetrg(pts, tri)
    tri_values = np.reshape(node_value[tri.ravel()], [m, -1])
    grad_el_x = np.sum(grad_phi_x * tri_values, axis=1)
    grad_el_y = np.sum(grad_phi_y * tri_values, axis=1)
    return grad_el_x, grad_el_y


def demo():
    """demo shows how to interpolate on regular/irregular grids"""
    # 1. create mesh
    mesh_obj, _ = layer_circle(n_layer=8, n_fan=6)
    pts = mesh_obj["node"]
    tri = mesh_obj["element"]

    # set anomaly
    anomaly = [{"x": 0.5, "y": 0.5, "d": 0.2, "perm": 100.0}]
    mesh_new = set_perm(mesh_obj, anomaly=anomaly)

    # 2. interpolate using averaged neighbor triangle area
    perm_node = sim2pts(pts, tri, mesh_new["perm"])

    # plot mesh and interpolated mesh (tri2pts)
    fig_size = (6, 4)
    fig = plt.figure(figsize=fig_size)
    ax = fig.add_subplot(111)
    ax.set_aspect("equal")
    ax.triplot(pts[:, 0], pts[:, 1], tri)
    im1 = ax.tripcolor(pts[:, 0], pts[:, 1], tri, mesh_new["perm"])
    fig.colorbar(im1, orientation="vertical")

    fig = plt.figure(figsize=fig_size)
    ax2 = fig.add_subplot(111)
    ax2.set_aspect("equal")
    ax2.triplot(pts[:, 0], pts[:, 1], tri)
    im2 = ax2.tripcolor(pts[:, 0], pts[:, 1], tri, perm_node, shading="flat")
    fig.colorbar(im2, orientation="vertical")

    # 3. interpolate on grids (irregular or regular) using IDW, sigmod
    xg, yg, mask = meshgrid(pts)
    im = np.ones_like(mask)
    # mapping from values on xy to values on xyi
    xy = np.mean(pts[tri], axis=1)
    xyi = np.vstack((xg.flatten(), yg.flatten())).T
    # w_mat = weight_idw(xy, xyi)
    w_mat = weight_sigmod(xy, xyi)
    im = np.dot(w_mat.T, mesh_new["perm"])
    # im = weight_linear_rbf(xy, xyi, mesh_new['perm'])
    im[mask] = 0.0
    # reshape to grid size
    im = im.reshape(xg.shape)

    # plot interpolated values
    fig, ax = plt.subplots(figsize=fig_size)
    ax.set_aspect("equal")
    ax.triplot(pts[:, 0], pts[:, 1], tri, alpha=0.5)
    im3 = ax.pcolor(xg, yg, im, edgecolors=None, linewidth=0, alpha=0.8)
    fig.colorbar(im3, orientation="vertical")
    plt.show()


if __name__ == "__main__":
    demo()



# coding: utf-8
# pylint: disable=invalid-name, no-member, too-many-locals
""" plot function for distmesh 2d and 3d """
from __future__ import absolute_import

import matplotlib
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np

from ..utils import edge_project, edge_list


def circumcircle(p1, p2, p3):
    """
    The circumcircle is a triangle's circumscribed circle,
    returns (x, y, r) of circumcenter

    Parameters
    ----------
    p1, p2, p3 : array_like
        points

    Note
    ----
    http://www.labri.fr/perso/nrougier/coding/gallery/
    """
    dp1 = p1 - p2
    dp2 = p3 - p1

    mid1 = (p1 + p2) / 2.0
    mid2 = (p3 + p1) / 2.0

    a = np.array([[-dp1[1], dp2[1]], [dp1[0], -dp2[0]]])
    b = -mid1 + mid2
    s = np.linalg.solve(a, b)
    # extract circumscribed center and radius
    cpc = mid1 + s[0] * np.array([-dp1[1], dp1[0]])
    cr = np.linalg.norm(p1 - cpc)

    return cpc[0], cpc[1], cr


def voronoi(pts, tri, fd=None):
    """
    build voronoi cells using delaunay tessellation

    Parameters
    ----------
    pts : array_like
        points on 2D
    tri : array_like
        triangle structure
    fd : str
        function handler of distances

    Returns
    -------
    array_like
        voronoi cells of lists

    Note
    ----
    adds 'edge-list using signed distance function'
    http://www.labri.fr/perso/nrougier/coding/gallery/
    """
    n = tri.shape[0]

    # Get circle for each triangle, center will be a voronoi cell point
    cells = []
    for i in range(pts.shape[0]):
        cells.append(list())

    def extract_xy(e):
        """append center (x,y) of triangle-circumcircle to the cell list"""
        p1, p2, p3 = pts[tri[e]]
        xc, yc, _ = circumcircle(p1, p2, p3)
        return [xc, yc]

    # list(map(extract_xy, range(n)))
    pc = np.array([extract_xy(i) for i in range(n)])

    # project point on the boundary if it is outside, where fd(p) > 0
    # this happens when low-quality mesh is generated.
    if fd is not None:
        d = fd(pc)
        ix = d > 0
        pc[ix] -= edge_project(pc[ix], fd)

    # build cells enclosing points
    for i in range(n):
        pc_tuple = tuple(pc[i])
        cells[tri[i, 0]].append(pc_tuple)
        cells[tri[i, 1]].append(pc_tuple)
        cells[tri[i, 2]].append(pc_tuple)

    # append middle (x, y) of edge-bars to the cells,
    # make a closed patch of the voronoi tessellation.
    # note : it may be better if you project this point on fd
    edge_bars = edge_list(tri)
    h_bars = np.mean(pts[edge_bars], axis=1)
    if fd is not None:
        h_bars -= edge_project(h_bars, fd)
    for i, bars in enumerate(edge_bars):
        cells[bars[0]].append(tuple(h_bars[i]))
        cells[bars[1]].append(tuple(h_bars[i]))

    x = pts[:, 0]
    y = pts[:, 1]
    # Reordering cell points in trigonometric way
    for i, cell in enumerate(cells):
        xy = np.array(cell)
        angles = np.arctan2(xy[:, 1] - y[i], xy[:, 0] - x[i])
        s = np.argsort(angles)
        cell = xy[s].tolist()
        cell.append(cell[0])
        cells[i] = cell

    return cells


def voronoi_plot(pts, tri, figsize=(6, 4), val=None, fd=None):
    """plot voronoi diagrams on bounded shape

    Parameters
    ----------
    pts : array_like
        points on 2D
    tri : array_like
        triangle structure
    val : array_like, optional
        values on nodes
    fd : str, optional
        function handler

    Returns
    -------
    fig : str
        figure handler
    ax : str
        axis handler

    Note
    ----
    adds 'maps value to colormap', see
    http://www.labri.fr/perso/nrougier/coding/gallery/
    """
    cells = voronoi(pts, tri, fd)

    # map values on nodes to colors
    if val is None:
        val = np.random.rand(pts.shape[0])
    norm = matplotlib.colors.Normalize(vmin=min(val), vmax=max(val), clip=True)
    mapper = cm.ScalarMappable(norm=norm, cmap=cm.Greens)

    fig, ax = plt.subplots(figsize=figsize)
    # draw mesh (optional)
    # ax.triplot(pts[:, 0], pts[:, 1], tri, color='b', alpha=0.50, lw=0.5)
    # ax.scatter(pts[:, 0], pts[:, 1], s=3, color='r', zorder=1)

    # draw voronoi tessellation
    for i, cell in enumerate(cells):
        codes = (
            [matplotlib.path.Path.MOVETO]
            + [matplotlib.path.Path.LINETO] * (len(cell) - 2)
            + [matplotlib.path.Path.CLOSEPOLY]
        )
        path = matplotlib.path.Path(cell, codes)
        # map values on nodes to colormap
        # e.g., color = np.random.uniform(.4, .9, 3)
        color = mapper.to_rgba(val[i])
        # using patches to plot the voronoi of a node
        patch = matplotlib.patches.PathPatch(
            path, facecolor=color, edgecolor="w", zorder=-1, lw=0.4
        )
        ax.add_patch(patch)

    ax.set_aspect("equal")
    return fig, ax



# pylint: disable=no-member, invalid-name
""" extract timestamp from file or path """
# Copyright (c) Benyuan Liu. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
import re
import pandas as pd


def string_to_time(fstr):
    """
    extract the start time from the path of .et3, .erd.
    """
    pattern = re.compile(r"(?P<date>\d+-\d+-\d+-\d+-\d+-\d+)")
    t = pattern.search(fstr)

    time_start = None
    if t is not None:
        time_start = t.group("date")
        # replace DATA2014-12-28-03-06-11 to DATA2014/12/28 03:06:11
        time_start = re.sub(r"-", r"/", time_start, count=2)
        time_start = re.sub(r"-", r" ", time_start, count=1)
        time_start = re.sub(r"-", r":", time_start)

    return time_start


def get_date_from_folder(file_str):
    """
    get datetime from file folder of .et3, .erd, i.e.,
    'DATA2015-01-29-16-57-30/', '2020-10-11-03-48-52/'
    """
    f = file_str.strip()
    f = f[:-1]  # remove trailing '/'
    f = f.replace("DATA", "")
    # replace the 3rd occurrence of '-'
    w = [m.start() for m in re.finditer(r"-", f)][2]
    # before w do not change, after w, '-' -> ':'
    f = f[:w] + " " + f[w + 1 :].replace("-", ":")
    # now f becomes '2015-01-29 16:57:30'
    return pd.to_datetime(f)
